Imports system.Math

''' <summary>
''' Motion for Conservative TeleOp behavior
''' </summary>
''' <remarks></remarks>

Public Class ConservativeTeleOpMotion
    Inherits Motion

    Private AlertObstacle As Boolean = False
    Private AlertIncline As Boolean = False

    Public Sub New(ByVal control As MotionControl)
        MyBase.New(control)
    End Sub

    Protected Overrides Sub OnActivated()
        MyBase.OnActivated()
        Me.Control.Agent.Halt()
    End Sub

    Protected Overrides Sub ProcessSensorUpdate(ByVal sensor As Sensor)
        MyBase.ProcessSensorUpdate(sensor)

        If sensor.SensorType = LaserRangeSensor.SENSORTYPE_RANGESCANNER Then
            Me.ProcessLaserRangeData(DirectCast(sensor, LaserRangeSensor).PeekData)
        ElseIf sensor.SensorType = SonarSensor.SENSORTYPE_SONAR Then
            Me.ProcessSonarData(DirectCast(sensor, SonarSensor).CurrentData)
        ElseIf sensor.SensorType = InsSensor.SENSORTYPE_INS Then
            Me.ProcessInsData(DirectCast(sensor, InsSensor).CurrentData)
        ElseIf sensor.SensorType = GroundTruthSensor.SENSORTYPE_GROUNDTRUTH Then
            Me.ProcessGroundTruthData(DirectCast(sensor, GroundTruthSensor).CurrentData)
        End If
    End Sub

    Protected Overridable Sub ProcessLaserRangeData(ByVal laser As LaserRangeData)
        Dim minFrontRange As Double = Double.MaxValue
        For i As Integer = 80 To 100
            minFrontRange = Min(minFrontRange, laser.Range(i))
        Next

        Dim elapsedSinceLastAlert As TimeSpan = DateTime.Now - Me.Control.Agent.MotionTimeStamp

        If minFrontRange < 2.0 And elapsedSinceLastAlert.Seconds > 2 Then
            Me.Control.Agent.Halt()
            Me.AlertObstacle = True
            Dim alert As String = "LSR - OBS - " + minFrontRange.ToString
            Me.Control.Agent.NotifyAlertReceived(alert)
        End If
    End Sub

    Protected Overridable Sub ProcessSonarData(ByVal sonar As SonarData)
        Dim minFrontRange As Double = Double.MaxValue
        For i As Integer = 3 To 6
            minFrontRange = Min(minFrontRange, sonar.Front(i))
        Next

        Dim elapsedSinceLastAlert As TimeSpan = DateTime.Now - Me.Control.Agent.MotionTimeStamp

        'If minFrontRange < 2.0 And elapsedSinceLastAlert.Seconds > 2 Then
        ' Me.Control.Agent.Halt()
        ' Me.AlertObstacle = True
        ' Dim alert As String = "SNR - OBS - " + minFrontRange.ToString
        ' Me.Control.Agent.NotifyAlertReceived(alert)
        ' End If
    End Sub

    Protected Overridable Sub ProcessGroundTruthData(ByVal current_data As GroundTruthData)
    End Sub

    Protected Overridable Sub ProcessInsData(ByVal ins As InsData)
    End Sub

End Class
