Imports System.IO

''' <summary>
''' Will use a logfile to obtain sensor readings.
''' </summary>
''' <remarks></remarks>
Public Class LogDriver
    Inherits Driver


    Public Delegate Function ParseDelegate(ByVal reciever As Agent, ByVal line As String) As Boolean


    Public Sub New(ByVal agent As Agent, ByVal logfile As String, ByVal parseHandler As ParseDelegate)
        MyBase.New(agent)
        Me._LogFile = logfile
        Me._Parse = parseHandler

        Dim logext As String = Path.GetExtension(logfile)
        Dim insext As String = ".ins"
        Dim insfile As String = logfile.Substring(0, logfile.Length - logext.Length) & insext

        If File.Exists(insfile) Then
            Me._InsFile = insfile
        Else
            Me._InsFile = ""
        End If

    End Sub


    Private _LogFile As String
    Public ReadOnly Property LogFile() As String
        Get
            Return Me._LogFile
        End Get
    End Property


    Private _InsFile As String
    Public ReadOnly Property InsFile() As String
        Get
            Return Me._InsFile
        End Get
    End Property


    Private _Parse As ParseDelegate
    Protected ReadOnly Property Parse() As ParseDelegate
        Get
            Return Me._Parse
        End Get
    End Property

    Protected Overrides Sub Run()

        System.Threading.Thread.CurrentThread.Name = Me.Agent.Name

        'open a reader to the logfile and optionally the odofile
        Dim logreader As StreamReader = File.OpenText(Path.GetFullPath(Me.LogFile))
        Dim insreader As StreamReader = Nothing
        If Not String.IsNullOrEmpty(Me.InsFile) Then
            insreader = File.OpenText(Path.GetFullPath(Me.InsFile))
        End If

        Dim logline As String
        Dim logtime As Double

        Dim insline As String = ""
        Dim instime As Double = -1

        'a single line will hold a single server message or sensor reading
        While Me.IsRunning AndAlso Not logreader.EndOfStream

            Me.CheckForPause()

            logline = logreader.ReadLine

            'skip commented lines
            If logline.StartsWith(My.Settings.CommentChar) AndAlso Not Me.Agent.AgentConfig.LogFileFormat = "Longwood" Then
                Continue While
            End If

            'skip INFO message, otherwise the Agent code will think that
            'the agent was actually spawned in UsarSim, would cause numerous
            'exceptions
            If logline.StartsWith("NFO") Then
                Continue While
            End If


            'ok, we found a line with sensor readings for the agent, now check if we need
            'to get odometry data first
            If Not IsNothing(insreader) Then

                'assume timestamps are in msecs and are the first double in the line
                logtime = Double.Parse(logline.Substring(0, logline.IndexOf(" ")))

                'see if we need to move forward in ins log
                If instime < 0 OrElse instime < logtime Then

                    If instime > 0 AndAlso Not String.IsNullOrEmpty(insline) Then
                        'forward cached reading first
                        If Not LineParsers.ParseInsLine(Me.Agent, insline) Then
                            Console.Error.WriteLine("Could not parse ins-line: " & vbNewLine & vbTab & insline)
                        End If
                    End If

                    'move forward in ins log
                    While Not insreader.EndOfStream
                        insline = insreader.ReadLine

                        'skip commented lines
                        If insline.StartsWith(My.Settings.CommentChar) Then
                            Continue While
                        End If

                        'get ins timestamp
                        instime = Double.Parse(insline.Substring(0, insline.IndexOf(" ")))
                        If instime < logtime Then
                            'forward INS reading to agent
                            If Not LineParsers.ParseInsLine(Me.Agent, insline) Then
                                Console.Error.WriteLine("Could not parse ins-line: " & vbNewLine & vbTab & insline)
                            End If
                        Else
                            'ok, ins is in sync with logtime now
                            'insline will cache the last INS reading
                            Exit While
                        End If

                    End While

                End If

            End If


            'forward line the the referenced parsing function
            If Not Me.Parse(Me.Agent, logline) Then
                Console.Error.WriteLine("Could not parse log-line: " & vbNewLine & vbTab & logline)
            End If

        End While

    End Sub

End Class
