Imports UvARescue.Agent
Imports UvARescue.Slam
Imports UvARescue.Tools

Imports System.IO
Imports System.Windows.Forms

Public Class AgentConfigDialog
    Inherits ConfigDialog


    Protected Enum AllowedRobotModels

        ComStation

        Talon
        P2AT
        Zerg

        ATRVJr
        Hummer
        P2DX
        OmniP2DX
        Telemax

        AirRobot

        AIBO

    End Enum

    Protected Enum SupportedScanMatchers

        WSM
        IDC
        MbICP

        MultiICP

        QuadWSM
        QuadIDC

    End Enum

    Protected Enum SupportedLogFileFormats

        UsarSim
        Player
        Carmen
        Cogniron
        Matlab
        NomadWithHokuyo
        Longwood

    End Enum

    Protected Enum SkinDetectorMode
        Detection
        Teaching
    End Enum

    Protected Enum BehaviorMode
        ' see BehaviorControl.BehaviorType
        TeleOperation
        ConservativeTeleOp
        AutonomousExploration
        DriveSquare
        ' RIKNOTE: added for assignment 2 and 3
        FollowCorridorBehaviorType
        ObstacleAvoidanceBehaviorType
    End Enum

    Public Sub New(ByVal config As AgentConfig)
        MyBase.New(config)

        ' This call is required by the Windows Form Designer.
        InitializeComponent()

        'load enums into combos
        ' RIKNOTE: converts the BehaviorMode enums to strings in the AgentConfig pulldown menu
        Me.LoadEnumIntoComboBox(GetType(AllowedRobotModels), Me.cboModel)
        Me.LoadEnumIntoComboBox(GetType(BehaviorMode), Me.cboBehaviorMode)
        Me.LoadEnumIntoComboBox(GetType(SupportedLogFileFormats), Me.cboLogfileFormat)
        Me.LoadEnumIntoComboBox(GetType(MappingMode), Me.cboMappingMode)
        Me.LoadEnumIntoComboBox(GetType(SupportedScanMatchers), Me.cboScanMatcher)
        Me.LoadEnumIntoComboBox(GetType(PoseEstimationSeedMode), Me.cboSeedMode)
        Me.LoadEnumIntoComboBox(GetType(SkinDetectorMode), Me.cboSkinDetMode)
        Me.LoadEnumIntoComboBox(GetType(ImageAnalysis.SkinDetectorTeacherMode), Me.cboTeachMode)

    End Sub


    Private Sub LoadEnumIntoComboBox(ByVal enumType As Type, ByVal combo As ComboBox)

        Dim names As Array = System.Enum.GetNames(enumType)
        Dim values As Array = System.Enum.GetValues(enumType)

        Dim table As New DataTable
        table.Columns.Add("Name", GetType(String))
        table.Columns.Add("Value", enumType)

        For i As Integer = 0 To names.Length - 1
            table.Rows.Add(names.GetValue(i), values.GetValue(i))
        Next

        With combo
            .DataSource = table
            .DisplayMember = "Name"
            .ValueMember = "Value"
        End With

    End Sub

    Private Sub ShowEnumValueInComboBox(ByVal enumType As Type, ByVal valueToSelect As String, ByVal combo As ComboBox)
        Dim values As Array = System.Enum.GetValues(enumType)
        For Each value As Object In values
            If value.ToString = valueToSelect Then
                combo.SelectedValue = value
                Exit For
            End If
        Next
    End Sub

    Public ReadOnly Property AgentConfig() As AgentConfig
        Get
            Return DirectCast(MyBase.Config, AgentConfig)
        End Get
    End Property


    Protected Overrides Sub OnOpened(ByVal filename As String)
        MyBase.OnOpened(filename)
        My.Settings.LastAgentConfigFile = filename
        My.Settings.Save()
    End Sub
    Protected Overrides Sub OnSaved(ByVal filename As String)
        MyBase.OnSaved(filename)
        My.Settings.LastAgentConfigFile = filename
        My.Settings.Save()
    End Sub


    Protected Overrides Sub SyncGuiWithConfig()
        MyBase.SyncGuiWithConfig()
        With Me.AgentConfig

            Me.Text = String.Format("Agent Configuration - [{0}]", .AgentName)

            Me.ShowEnumValueInComboBox(GetType(AllowedRobotModels), .RobotModel, Me.cboModel)

            Me.txtLocation.Text = .StartLocation
            Me.txtRotation.Text = .StartRotation
            Me.chkSpawnFromCommander.Checked = .SpawnFromCommander
            Me.numAgentNumber.Value = .AgentNumber

            If Not String.IsNullOrEmpty(.AgentType) Then
                Select Case .AgentType.ToLower
                    Case "operator"
                        Me.optUsarOperator.Checked = True
                    Case "slam"
                        Me.optUsarSlam.Checked = True
                    Case "follow"
                        Me.optUsarFollow.Checked = True
                    Case "skin"
                        Me.optUsarSkinDet.Checked = True
                    Case Else
                        Me.optUsarBase.Checked = True
                End Select
            End If

            Me.optPlayback.Checked = .LogPlayback
            Me.lnkLogfile.Text = .LogFile
            Me.ShowEnumValueInComboBox(GetType(SupportedLogFileFormats), .LogFileFormat, Me.cboLogfileFormat)

            Me.optLive.Checked = Not .LogPlayback
            Me.chkUseLogger.Checked = .UseLogger
            Me.chkUseImageServer.Checked = .UseImageServer
            Me.chkUseMultiView.Checked = .UseMultiView
            Me.txtMultiViewPanels.Text = .MultiViewPanels

            Me.ShowEnumValueInComboBox(GetType(BehaviorMode), .BehaviorMode, Me.cboBehaviorMode)
            Me.txtBehaviorBalance.Text = CStr(.BehaviorBalance)
            Me.ShowEnumValueInComboBox(GetType(MappingMode), .MappingMode, Me.cboMappingMode)
            Me.ShowEnumValueInComboBox(GetType(SupportedScanMatchers), .ScanMatcher, Me.cboScanMatcher)
            Me.ShowEnumValueInComboBox(GetType(PoseEstimationSeedMode), .SeedMode, Me.cboSeedMode)
            Me.chkNoise.Checked = .UseNoise
            Me.txtNoiseSigma.Text = CStr(.NoiseSigma)

            Me.ShowEnumValueInComboBox(GetType(SkinDetectorMode), .SkinDetectorMode, Me.cboSkinDetMode)
            Me.ShowEnumValueInComboBox(GetType(ImageAnalysis.SkinDetectorTeacherMode), .TeacherMode, Me.cboTeachMode)

            Me.txtDetTheta.Text = .DetectorTheta
        End With
    End Sub

    Protected Overrides Sub SyncConfigWithGui()
        MyBase.SyncConfigWithGui()
        With Me.AgentConfig

            If Me.optUsarOperator.Checked Then
                .AgentType = "operator"
            ElseIf Me.optUsarSlam.Checked Then
                .AgentType = "slam"
            ElseIf Me.optUsarSkinDet.Checked Then
                .AgentType = "skin"
            ElseIf Me.optUsarFollow.Checked Then
                .AgentType = "follow"
            Else
                .AgentType = "base"
            End If

            .RobotModel = CType(Me.cboModel.SelectedValue, AllowedRobotModels).ToString
            .StartLocation = Me.txtLocation.Text
            .StartRotation = Me.txtRotation.Text
            .SpawnFromCommander = Me.chkSpawnFromCommander.Checked
            .AgentNumber = CInt(Me.numAgentNumber.Value)

            .LogPlayback = Me.optPlayback.Checked
            .LogFile = Me.lnkLogfile.Text
            .LogFileFormat = CType(Me.cboLogfileFormat.SelectedValue, SupportedLogFileFormats).ToString

            .UseLogger = Me.chkUseLogger.Checked
            .UseImageServer = Me.chkUseImageServer.Checked
            .UseMultiView = Me.chkUseMultiView.Checked
            .MultiViewPanels = Me.txtMultiViewPanels.Text

            .MappingMode = CType(Me.cboMappingMode.SelectedValue, MappingMode).ToString
            .BehaviorMode = CType(Me.cboBehaviorMode.SelectedValue, BehaviorMode).ToString
            If IsNumeric(Me.txtBehaviorBalance.Text) Then
                .BehaviorBalance = Double.Parse(Me.txtBehaviorBalance.Text)
            End If

            .ScanMatcher = CType(Me.cboScanMatcher.SelectedValue, SupportedScanMatchers).ToString
            .SeedMode = CType(Me.cboSeedMode.SelectedValue, PoseEstimationSeedMode).ToString
            .UseNoise = Me.chkNoise.Checked
            If IsNumeric(Me.txtNoiseSigma.Text) Then
                .NoiseSigma = Double.Parse(Me.txtNoiseSigma.Text)
            End If

            .SkinDetectorMode = CType(Me.cboSkinDetMode.SelectedValue, SkinDetectorMode).ToString
            .TeacherMode = CType(Me.cboTeachMode.SelectedValue, ImageAnalysis.SkinDetectorTeacherMode).ToString
            .DetectorTheta = Me.txtDetTheta.Text
        End With
    End Sub


    Private Sub optUsarOperator_CheckedChanged(ByVal sender As Object, ByVal e As System.EventArgs) Handles optUsarOperator.CheckedChanged
        If Me.optUsarOperator.Checked Then
            Me.chkSpawnFromCommander.Checked = True
        End If
    End Sub


    Private Sub optUsarSlam_CheckedChanged(ByVal sender As Object, ByVal e As System.EventArgs) Handles optUsarSlam.CheckedChanged
        Me.grpSlamConfig.Enabled = Me.optUsarSlam.Checked OrElse Me.optUsarSkinDet.Checked
    End Sub

    Private Sub optUsarSkinDet_CheckedChanged(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles optUsarSkinDet.CheckedChanged
        Me.grpSlamConfig.Enabled = Me.optUsarSlam.Checked OrElse Me.optUsarSkinDet.Checked
        Me.grpSkinDet.Enabled = Me.optUsarSkinDet.Checked
        If Me.txtDetTheta.Text = "" Then
            Me.txtDetTheta.Text = "50"
        End If
    End Sub



    Private Sub optPlayback_CheckedChanged(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles optPlayback.CheckedChanged
        Me.pnlPlayback.Enabled = Me.optPlayback.Checked
    End Sub

    Private Sub optLive_CheckedChanged(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles optLive.CheckedChanged
        Me.pnlLive.Enabled = Me.optLive.Checked
    End Sub

    Private Sub lnkLogfile_LinkClicked(ByVal sender As System.Object, ByVal e As System.Windows.Forms.LinkLabelLinkClickedEventArgs) Handles lnkLogfile.LinkClicked
        Dim dlg As New System.Windows.Forms.OpenFileDialog
        Select Case dlg.ShowDialog
            Case Windows.Forms.DialogResult.OK
                Me.lnkLogfile.Text = dlg.FileName
        End Select
    End Sub
    Private Sub lnkLogfile_TextChanged(ByVal sender As Object, ByVal e As System.EventArgs) Handles lnkLogfile.TextChanged
        If String.IsNullOrEmpty(Me.lnkLogfile.Text) Then
            Me.lnkLogfile.Text = "<Click to Select ...>"
        End If
    End Sub

    Private Sub chkUseImageServer_CheckedChanged(ByVal sender As Object, ByVal e As System.EventArgs) Handles chkUseImageServer.CheckedChanged
        Me.pnlMultiView.Enabled = Me.chkUseImageServer.Checked
    End Sub

    Private Sub chkUseMultiView_CheckedChanged(ByVal sender As Object, ByVal e As System.EventArgs) Handles chkUseMultiView.CheckedChanged
        Me.lblMultiViewPanels.Enabled = Me.chkUseMultiView.Checked
        Me.txtMultiViewPanels.Enabled = Me.chkUseMultiView.Checked
    End Sub

    Private Sub chkNoise_CheckedChanged(ByVal sender As Object, ByVal e As System.EventArgs) Handles chkNoise.CheckedChanged
        Me.txtNoiseSigma.Enabled = Me.chkNoise.Checked
    End Sub

    Private Sub Label1_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles lblBehaviorBalance.Click

    End Sub

    Private Sub cboBehaviorMode_SelectedIndexChanged(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles cboBehaviorMode.SelectedIndexChanged

    End Sub
End Class
